/*-
 * Copyright (c) 1998-2010 Joe Marcus Clarke <marcus@marcuscom.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Gone - a terminal locking program with many improvements over lock(1)
 *
 * $MCom: gone/gone.c,v 1.10 2010/08/01 16:09:18 marcus Exp $
 */

#include "config.h"
#include <stdlib.h> /*  atoi() */
#include <stdio.h> /*  fprintf(), stderr, printf(), fgets(), stdin, BUFSIZ, fflush(), putchar(), clearerr() */
#include <signal.h> /*  signal(), SIGINT, SIGSTP, SIGQUIT, SIGALRM, SIGHUP, kill() */
#include <unistd.h> /*  getpass(), getuid(), getppid(), STDIN_FILENO, setuid() */
#include <ctype.h> /*  isdigit() */
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h> /*  struct timeval, struct itimerval, chmod(), gettimeofday(), localtime(), asctime(), setitimer(), stat(), struct stat */
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_SYS_TERMIO_H
#include <sys/termio.h> /*  struct termio, TCGETA, ECHO, TCSETAF, TCSETA, TCGETA, ioctl() */
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_SYS_CONSIO_H
#include <sys/consio.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#if defined(OSF1) && defined(EN_SEC)
#include <sys/security.h>
#include <prot.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#include "gone.h"

#define TIMEOUT 10
#ifdef _PASSWORD_LEN
#define PASSWORD_BUFFER _PASSWORD_LEN
#else
#define PASSWORD_BUFFER 128
#endif

#if defined OSF1 || defined ULTRIX
struct timeval  timeout = {0, 0};
struct timeval  zerotime = {0, 0};
#else /* OSF1 || ULTRIX */
struct timeval timeout;
struct timeval zerotime;
#endif /* OSF1 || ULTRIX */
#define SALT "3C" /* XXX Crappy */
struct stat ttystat;
long nexttime;
mode_t ttymode;
char *ttynam;
static int vtylocked = 0;

/*  Function prototypes */
void lockTerm(char *ap, char *password, int timeAway,
              int extendedCommands, int printBanner);
char *getPassword();
void usage(char *progName);
void quit();
void hi();
void printHeader(int extendedCommands, int printBanner);
void print_version_info(void);
void unlockVty(void);

struct gone_commands commands[] =
    {
	    {"w", "Print a summary of current system activity", 1},
	    {"finger", "Display user information", 1},
	    {"whoami", "Display the the locking user's name", 1},
	    {"help", "Print this message", 0},
	    {"date", "Display the date and time", 1},
	    {"status", "Display the lock timeout", 0},
	    {"cls", "Clear the screen, and display this message", 0}
    };

#if defined(ULTRIX)
extern char *optarg;
extern int optind, opterr;
#endif /* ULTRIX */

int
main(int argc, char *argv[]) {
#if !defined(HAVE_ERRNO_H)
    extern int errno;
#endif
    char *ap;
    char buffer[PASSWORD_BUFFER];
    char *cryptBuffer;
    char *cryptRetypeBuffer;
    char enteredPass[PASSWORD_BUFFER];
    struct timeval timval;
    struct itimerval ntimer, otimer;
    struct tm *timp;
    int goneTime;
    int ch;
    int mesgOff = 1;
    int extendedCommands = 1;
    int printBanner = 1;
    int lockvty = 0;
#if defined(SETUID)
    int useSysPasswd = 0;
    char *sysPass;
    struct passwd *pw;
#endif

    goneTime = TIMEOUT;
#if defined(OSF1) && defined(EN_SEC) && defined(SETUID)
    set_auth_parameters();
    while ((ch = getopt(argc, argv, "bvpxynlt:")) != -1)
#elif defined(SETUID)
    while ((ch = getopt(argc, argv, "bvpxynlt:")) != -1)
#else /* OSF1 && EN_SEC && SETUID */
    while ((ch = getopt(argc, argv, "bvyxnlt:")) != -1)
#endif /* OSF1 && EN_SEC && SETUID */
        switch((char)ch) {
        case 'b':
            printBanner = 0;
            break;
        case 'v':
            print_version_info();
            exit(0);
#if defined(OSF1) && defined(EN_SEC) && defined(SETUID)
        case 'p':
            if (getespwuid(getuid()) == NULL) {
                printf("Error getting your password from the system.\n");
                useSysPasswd = 0;
            }
            else {
                sysPass = getespwuid(getuid())->ufld->fd_encrypt;
                useSysPasswd = 1;
            }
            break;
#elif defined(SETUID)
        case 'p':
            if (!(pw = getpwuid(getuid()))) {
                printf("Error getting your password from the system.\n");
                useSysPasswd = 0;
            }
            else {
                sysPass = pw->pw_passwd;
                useSysPasswd = 1;
            }
            break;
#endif /* OSF1 && EN_SEC && SETUID */
        case 'n':
        case 'x':
            extendedCommands = 0;
            break;
        case 'y':
            mesgOff = 0;
            break;
	case 'l':
	    lockvty = 1;
	    break;
        case 't':
            if ((goneTime = atoi(optarg)) < 0) {
                /* If goneTime == 0, then lock the
                   terminal indifinitely. */
                fprintf(stderr, "Illegal timeout value.\n");
                exit(1);
            }
            break;
        default:
            usage(argv[0]);
        }

    setuid(getuid());  /* reset uid permissions */
    putenv("PATH=/bin:/usr/bin:/usr/bsd:/usr/ucb");

    if (!(ttynam = ttyname(STDIN_FILENO))) { /*  check to see if stdin is a tty */
        printf("Not a terminal\n");
        exit(1);
    }
    stat(ttynam, &ttystat); /*  save current tty permissions */
    ttymode = ttystat.st_mode;
    if (mesgOff == 1) {
        chmod(ttynam, 00600); /*  chmod the tty so only the owner can read and write to it */
    }
    else {
#ifdef IRIX
        /* Dumb IRIX "security" */
        chmod(ttynam, 00622);
#else /* IRIX */
        chmod(ttynam, 00620);
#endif /* IRIX */
    }

    timeout.tv_sec = goneTime * 60;

    if (gettimeofday(&timval,(struct timezone *)NULL)) {
        exit(1);
    }

    nexttime = timval.tv_sec + (goneTime * 60);
    timp = localtime((time_t *)&timval.tv_sec);
    ap = asctime(timp);


    ntimer.it_interval = zerotime;
    ntimer.it_value = timeout;

    if ( goneTime != 0 ) {
        /* Allow for the terminal to be locked forever. */
        setitimer(ITIMER_REAL, &ntimer, &otimer);
    }


    if (gettimeofday(&timval,(struct timezone *)NULL)) {
        exit(1);
    }

    nexttime = timval.tv_sec + (goneTime * 60);
    timp = localtime((time_t *)&timval.tv_sec);
    ap = asctime(timp);

    ntimer.it_interval = zerotime;
    ntimer.it_value = timeout;

    if ( goneTime != 0 ) {
        /* Allow for the terminal to be locked forever. */
        setitimer(ITIMER_REAL, &ntimer, &otimer);
    }

#if defined(SETUID)
    if (useSysPasswd) {
        strncpy(enteredPass,sysPass, PASSWORD_BUFFER);
    } else
#endif
    {
        strncpy(buffer, getpass("Password: "), PASSWORD_BUFFER);

        if (*buffer == '\0') {
            printf("Empty passwords are not allowed!\n");
            exit(1);
        }
        cryptBuffer = (char *)crypt(buffer, SALT);
        strncpy(enteredPass, cryptBuffer, PASSWORD_BUFFER);
	memset(buffer, '\0', sizeof(buffer));

        strncpy(buffer, getpass("Retype: "), PASSWORD_BUFFER);
        cryptRetypeBuffer = (char *)crypt(buffer, SALT);
	memset(buffer, '\0', sizeof(buffer));
        fflush(stdin);

        if (strcmp(enteredPass, cryptRetypeBuffer)) {
            printf("Passwords did not match!\n");
            exit(1);
        }
    }
    /* Lock the vty if possible */
#ifdef VT_LOCKSWITCH
    if (lockvty) {
        if (ioctl(0, VT_LOCKSWITCH, &lockvty) == -1) {
	    printf("Failed to prevent VTY switching: %s\n", strerror(errno));
	    exit(1);
        }
        vtylocked = 1;
    }
#else
    printf("VTY locking is not supported on this platform.\n");
#endif

    /* set signal handlers */
    signal(SIGINT, (void (*)(int))hi);
    signal(SIGQUIT, (void (*)(int))hi);
    signal(SIGTSTP, (void (*)(int))hi);
    signal(SIGALRM, (void (*)(int))quit);
    signal(SIGHUP, (void (*)(int))quit);
    lockTerm(ap, enteredPass, goneTime, extendedCommands, printBanner);

    return 0;
}

void lockTerm(char *ap, char *password, int timeAway,
              int extendedCommands, int printBanner) {
    char key[PASSWORD_BUFFER];
    char *cryptKey;
    char cmd[BUFSIZ];
    int badPass = 1;
    int goodCommand;
    int i;


    if ( timeAway == 0 ) {
        printf("Terminal locked at %s indefinitely.\n", ap);
    }
    else {
        printf("Terminal locked at %s for %d minutes(s).\n", ap, timeAway);
    }

    printHeader(extendedCommands, printBanner);
    fflush(stdin);
    printf("> ");

    while(badPass) {
        fflush(stdin);
        if (!fgets(cmd, sizeof(cmd), stdin)) {
            clearerr(stdin);
            continue;
        }
        goodCommand = 0;
	cmd[strlen(cmd) - 1] = '\0';
        if (!strcmp(cmd, "help")) {
            printHeader(extendedCommands, printBanner);
        }
        else if (!strcmp(cmd, "cls")) {
            system("clear");
            printHeader(extendedCommands, printBanner);
        }
        else if (!strcmp(cmd, "status")) {
            hi();
        }
        else if (extendedCommands && !strcmp(cmd, "finger")) {
            char fingerCommand[129] = "finger ";
            char hostname[120];
            char parsedHostname[120];
	    memset(hostname, '\0', sizeof(hostname));
	    memset(parsedHostname, '\0', sizeof(parsedHostname));
            printf("Enter parameters: ");
            if (!fgets(hostname, sizeof(hostname), stdin)) {
                clearerr(stdin);
                continue;
            }
            sscanf(hostname, "%[^]^#~()[\\\x0A\xFF\"'|`<>*?&$;,]", parsedHostname);
            strncat(fingerCommand, parsedHostname, sizeof(fingerCommand) - 1);
            system(fingerCommand);
        }

        else {
            for (i = 0; i < sizeof(commands)/sizeof(commands[0]); i++) {
                if (extendedCommands && !strcmp(cmd, commands[i].command)) {
                    system(commands[i].command);
                    goodCommand = 1;
                }
            }
        }
        if (!goodCommand && strcmp(cmd, "unlock") && strcmp(cmd, "") && strcmp(cmd, "help") && strcmp(cmd, "cls") && (!extendedCommands || strcmp(cmd, "finger")) && strcmp(cmd, "status")) {
            printf("%s: command not allowed in lock mode\n", cmd);
            printf("Type 'help' for a list of commands.\n");
        }
        if (strcmp(cmd, "unlock") || cmd[0] == '\0') printf("> ");
        else {
            strncpy(key, getpass("Enter password: "), PASSWORD_BUFFER);
#if defined(OSF1) && defined(EN_SEC) && defined(SETUID)
            cryptKey = (char *)bigcrypt(key, password);
#else /* OSF1 && EN_SEC && SETUID */
            cryptKey = (char *)crypt(key, password);
#endif /* OSF1 && EN_SEC && SETUID */
            if (strcmp(cryptKey, password)) printf("Incorrect password!\n> ");
            else {
                chmod(ttynam, ttymode); /*  reset tty permissions */
                badPass = 0;
            }
        }
    }

    unlockVty();

    exit(0);

}


void usage(char *progName) {
#ifdef SETUID
    printf("Usage: %s [-v] [-p] [-x] [-l] [-y] [-t timeout] [Gone message]\n", progName);
#else /* SETUID */
    printf("Usage: %s [-v] [-x] [-l] [-y] [-t timeout] [Gone message]\n", progName);
#endif /* SETUID */
    exit(1);
}

void quit() {  /*  log users out by killing their shell */
    unlockVty();
    kill(getppid(), SIGKILL);
    exit(0);
}


void printHeader(int extendedCommands, int printBanner) {
    int i;

    if (printBanner) {
        printf("\n\n\n");
        printf("                   XXXXX  XX    X        X    X  XXXXX  XXXXX\n");
        printf("                     X    X X   X        X    X  X      X\n");
        printf("                     X    X  X  X        X    X  XXXXX  XXX\n");
        printf("                     X    X   X X        X    X      X  X\n");
        printf("                   XXXXX  X    XX        XXXXXX  XXXXX  XXXXX\n");
    }
    printf("\n");

    printf("Available commands: \n");
    for(i = 0; i < sizeof(commands)/sizeof(commands[0]); i++) {
        if (extendedCommands || !commands[i].extended) {
            printf("\t%-8s : %s\n", commands[i].command, commands[i].help_str);
        }
    }
    printf("\n\tType unlock to unlock the terminal\n\n");
}

void
hi() {
    struct timeval timval;

#ifdef IRIX
    signal(SIGINT, (void (*)(int))hi);
    signal(SIGQUIT, (void (*)(int))hi);
    signal(SIGTSTP, (void (*)(int))hi);
    signal(SIGALRM, (void (*)(int))quit);
    signal(SIGHUP, (void (*)(int))quit);
#endif /* IRIX */

    if (!gettimeofday(&timval,(struct timezone *)NULL)) {
        printf("\nTerminal locked!\nType 'help' for a list of commands\n\n");
        if ( nexttime - timval.tv_sec >= 0 ) {
            printf("Timeout in %ld minutes and %ld seconds\n", (nexttime - timval.tv_sec)/60, (nexttime - timval.tv_sec)%60);
        }
        else {
            printf("Terminal locked indefinitely.\n");
            printf("Terminal has been locked for %ld minutes and %ld seconds.\n", (timval.tv_sec - nexttime)/60, (timval.tv_sec - nexttime)%60);
        }
    }
}

void print_version_info(void) {
    printf("Gone version %s by Joe Marcus Clarke.\n", VERSION);
    printf("\t* Gone compiled on %s at %s for %s.\n", __DATE__, __TIME__, OSTYPE);
#if defined(SETUID)
    printf("\t* Gone was installed setuid to root.\n");
#endif /* SETUID */
#if defined(OSF1) && (EN_SEC) 
    printf("\t* Gone is OSF/1 Enhanced Security-aware.\n");
#endif /* OSF1 && EN_SEC */
    printf("\t* Gone copyright MarcusCom 1998-2010.\n");
}

void unlockVty() {
#ifdef VT_LOCKSWITCH
    if (vtylocked) {
  	int unlock = 0x2;
	(void)ioctl(0, VT_LOCKSWITCH, &unlock);
    }
#endif
}

